"""
The MIT License (MIT)

Copyright (c) 2021-present Dolfies

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
"""

from __future__ import annotations

from datetime import datetime, timedelta
from typing import TYPE_CHECKING, List, Optional, Union

from .enums import PaymentSourceType, PromotionType, SubscriptionInterval, try_enum
from .flags import PromotionFlags
from .mixins import Hashable
from .subscriptions import SubscriptionTrial
from .utils import _get_as_snowflake, parse_time, utcnow

if TYPE_CHECKING:
    from .state import ConnectionState
    from .types.promotions import (
        ClaimedPromotion as ClaimedPromotionPayload,
        Promotion as PromotionPayload,
        TrialOffer as TrialOfferPayload,
        DiscountOffer as DiscountOfferPayload,
        Discount as DiscountPayload,
        UserOffer as UserOfferPayload,
        PricingPromotion as PricingPromotionPayload,
    )

__all__ = (
    'Promotion',
    'UserOffer',
    'TrialOffer',
    'DiscountOffer',
    'Discount',
    'PricingPromotion',
)


class Promotion(Hashable):
    """Represents a Discord promotion.

    .. container:: operations

        .. describe:: x == y

            Checks if two promotions are equal.

        .. describe:: x != y

            Checks if two promotions are not equal.

        .. describe:: hash(x)

            Returns the promotion's hash.

        .. describe:: str(x)

            Returns the outbound promotion's name.

    .. versionadded:: 2.0

    Attributes
    ----------
    id: :class:`int`
        The promotion ID.
    type: :class:`PromotionType`
        The type of promotion.

        .. versionadded:: 2.1
    trial_id: Optional[:class:`int`]
        The trial ID of the inbound promotion, if applicable.
    starts_at: :class:`datetime.datetime`
        When the promotion starts.
    ends_at: :class:`datetime.datetime`
        When the promotion ends.
    claimed_at: Optional[:class:`datetime.datetime`]
        When the promotion was claimed.
        Only available for claimed promotions.
    code: Optional[:class:`str`]
        The promotion's claim code. Only available for claimed promotions.
    outbound_title: :class:`str`
        The title of the outbound promotion.
    outbound_description: :class:`str`
        The description of the outbound promotion.
    outbound_link: :class:`str`
        The redemption page of the outbound promotion, used to claim it.
    outbound_redemption_ends_at: Optional[:class:`datetime.datetime`]
        The end date of the outbound promotion's redemption period.

        .. versionadded:: 2.1
    outbound_restricted_countries: List[:class:`str`]
        The countries that the outbound promotion is not available in.
    inbound_title: Optional[:class:`str`]
        The title of the inbound promotion. This is usually Discord Nitro.
    inbound_description: Optional[:class:`str`]
        The description of the inbound promotion.
    inbound_link: Optional[:class:`str`]
        The Discord help center link of the inbound promotion.
    inbound_restricted_countries: List[:class:`str`]
        The countries that the inbound promotion is not available in.
    terms_and_conditions: :class:`str`
        The terms and conditions of the promotion.
    """

    __slots__ = (
        'id',
        'type',
        'trial_id',
        'starts_at',
        'ends_at',
        'claimed_at',
        'code',
        'outbound_title',
        'outbound_description',
        'outbound_link',
        'outbound_redemption_ends_at',
        'outbound_restricted_countries',
        'inbound_title',
        'inbound_description',
        'inbound_link',
        'inbound_restricted_countries',
        'terms_and_conditions',
        '_flags',
        '_state',
    )

    def __init__(self, *, data: Union[PromotionPayload, ClaimedPromotionPayload], state: ConnectionState) -> None:
        self._state = state
        self._update(data)

    def __str__(self) -> str:
        return self.outbound_title

    def __repr__(self) -> str:
        return f'<Promotion id={self.id} title={self.outbound_title!r}>'

    def _update(self, data: Union[PromotionPayload, ClaimedPromotionPayload]) -> None:
        promotion: PromotionPayload = data.get('promotion', data)

        self.id: int = int(promotion['id'])
        self.type = try_enum(PromotionType, promotion.get('type', 0))
        self.trial_id: Optional[int] = _get_as_snowflake(promotion, 'trial_id')
        self.starts_at: datetime = parse_time(promotion['start_date'])
        self.ends_at: datetime = parse_time(promotion['end_date'])
        self.claimed_at: Optional[datetime] = parse_time(data.get('claimed_at'))
        self.code: Optional[str] = data.get('code')
        self._flags: int = promotion.get('flags', 0)

        self.outbound_title: str = promotion.get('outbound_title', '')
        self.outbound_description: str = promotion.get('outbound_redemption_modal_body', '')
        self.outbound_link: str = promotion.get(
            'outbound_redemption_page_link',
            promotion.get('outbound_redemption_url_format', '').replace('{code}', self.code or '{code}'),
        )
        self.outbound_redemption_ends_at: Optional[datetime] = parse_time(promotion.get('outbound_redemption_end_date'))
        self.outbound_restricted_countries: List[str] = promotion.get('outbound_restricted_countries', [])
        self.inbound_title: Optional[str] = promotion.get('inbound_header_text')
        self.inbound_description: Optional[str] = promotion.get('inbound_body_text')
        self.inbound_link: Optional[str] = promotion.get('inbound_help_center_link')
        self.inbound_restricted_countries: List[str] = promotion.get('inbound_restricted_countries', [])
        self.terms_and_conditions: str = promotion.get('outbound_terms_and_conditions', '')

    @property
    def flags(self) -> PromotionFlags:
        """:class:`PromotionFlags`: Returns the promotion's flags."""
        return PromotionFlags._from_value(self._flags)

    def is_claimed(self) -> bool:
        """:class:`bool`: Checks if the promotion has been claimed.

        Only accurate if the promotion was fetched from :meth:`Client.promotions` with ``claimed`` set to ``True`` or :meth:`claim` was just called.
        """
        return self.claimed_at is not None

    def is_active(self) -> bool:
        """:class:`bool`: Checks if the promotion is active."""
        return self.starts_at <= utcnow() <= self.ends_at

    async def claim(self) -> str:
        """|coro|

        Claims the promotion.

        Sets :attr:`claimed_at` and :attr:`code`.

        Raises
        ------
        Forbidden
            You are not allowed to claim the promotion.
        HTTPException
            Claiming the promotion failed.

        Returns
        -------
        :class:`str`
            The claim code for the outbound promotion.
        """
        data = await self._state.http.claim_promotion(self.id)
        self._update(data)
        return data['code']


class UserOffer:
    """Represents a Discord user offer.

    .. versionadded:: 2.1

    Attributes
    ----------
    trial_offer: Optional[:class:`TrialOffer`]
        The trial offer.
    discount_offer: Optional[:class:`DiscountOffer`]
        The discount offer.
    discount: Optional[:class:`DiscountOffer`]
        The discount applied.
    """

    __slots__ = (
        'trial_offer',
        'discount_offer',
        'discount',
        '_state',
    )

    def __init__(self, *, data: UserOfferPayload, state: ConnectionState) -> None:
        self._state = state
        self.trial_offer: Optional[TrialOffer] = None
        self.discount_offer: Optional[DiscountOffer] = None
        self.discount: Optional[DiscountOffer] = None

        self._update(data)

    def _update(self, data: UserOfferPayload) -> None:
        state = self._state

        # Avoid nulling out fields through partial updates
        trial_offer = data.get('user_trial_offer')
        if trial_offer is not None:
            self.trial_offer = TrialOffer(data=trial_offer, state=state)

        discount_offer = data.get('user_discount_offer')
        if discount_offer is not None:
            self.discount_offer = DiscountOffer(data=discount_offer, state=state)

        discount = data.get('user_discount')
        if discount is not None:
            self.discount = DiscountOffer(data=discount, state=state)

    async def ack(self) -> None:
        """|coro|

        Acknowledges both the trial and discount offers.

        Raises
        ------
        NotFound
            The offers were not found.
        HTTPException
            Acknowledging the offers failed.
        """
        # The client sets all offers to null if this 404s
        # Unsure if I want to do that here
        data = await self._state.http.ack_user_offer(
            trial_offer_id=self.trial_offer.id if self.trial_offer else None,
            discount_offer_id=self.discount_offer.id if self.discount_offer else None,
        )
        if data:
            self._update(data)


class TrialOffer(Hashable):
    """Represents a Discord user trial offer.

    .. container:: operations

        .. describe:: x == y

            Checks if two trial offers are equal.

        .. describe:: x != y

            Checks if two trial offers are not equal.

        .. describe:: hash(x)

            Returns the trial offer's hash.

    .. versionadded:: 2.0

    Attributes
    ----------
    id: :class:`int`
        The ID of the trial offer.
    expires_at: Optional[:class:`datetime.datetime`]
        When the trial offer expires, if it has been acknowledged.
    trial_id: :class:`int`
        The ID of the trial.
    trial: :class:`SubscriptionTrial`
        The trial offered.
    """

    __slots__ = (
        'id',
        'expires_at',
        'trial_id',
        'trial',
        '_state',
    )

    def __init__(self, *, data: TrialOfferPayload, state: ConnectionState) -> None:
        self._state = state
        self._update(data)

    def _update(self, data: TrialOfferPayload) -> None:
        self.id: int = int(data['id'])
        self.expires_at: Optional[datetime] = parse_time(data.get('expires_at'))
        self.trial_id: int = int(data['trial_id'])
        self.trial: SubscriptionTrial = SubscriptionTrial(data['subscription_trial'])

    def __repr__(self) -> str:
        return f'<TrialOffer id={self.id} trial={self.trial!r}>'

    def is_acked(self) -> bool:
        """:class:`bool`: Checks if the trial offer has been acknowledged.

        .. versionadded:: 2.1
        """
        return self.expires_at is not None

    async def ack(self) -> None:
        """|coro|

        Acknowledges the trial offer.

        Raises
        ------
        HTTPException
            Acknowledging the trial offer failed.
        """
        data = await self._state.http.ack_user_offer(trial_offer_id=self.id)
        if not data:
            return

        # The type checker has no idea what is going on here
        if data.get('user_trial_offer') and int(data['user_trial_offer']['id']) == self.id:  # type: ignore
            self._update(data['user_trial_offer'])  # type: ignore


class DiscountOffer(Hashable):
    """Represents a Discord user discount offer.

    .. container:: operations

        .. describe:: x == y

            Checks if two discount offers are equal.

        .. describe:: x != y

            Checks if two discount offers are not equal.

        .. describe:: hash(x)

            Returns the discount offer's hash.

    .. versionadded:: 2.1

    Attributes
    ----------
    id: :class:`int`
        The ID of the discount offer.
    expires_at: Optional[:class:`datetime.datetime`]
        When the discount offer expires, if it has been acknowledged.
    applied_at: Optional[:class:`datetime.datetime`]
        When the discount offer was applied.
    deleted_at: Optional[:class:`datetime.datetime`]
        When the discount offer was deleted.
    invoice_id: Optional[Snowflake]
        The ID of the invoice the discount was applied to.
    discount_id: :class:`int`
        The ID of the discount.
    discount: :class:`Discount`
        The discount offered.
    """

    __slots__ = (
        'id',
        'expires_at',
        'applied_at',
        'deleted_at',
        'invoice_id',
        'discount_id',
        'discount',
        '_state',
    )

    def __init__(self, *, data: DiscountOfferPayload, state: ConnectionState) -> None:
        self._state = state
        self._update(data)

    def _update(self, data: DiscountOfferPayload) -> None:
        self.id: int = int(data['id'])
        self.expires_at: Optional[datetime] = parse_time(data.get('expires_at'))
        self.applied_at: Optional[datetime] = parse_time(data.get('applied_at'))
        self.deleted_at: Optional[datetime] = parse_time(data.get('deleted_at'))
        self.invoice_id: Optional[int] = _get_as_snowflake(data, 'invoice_id')
        self.discount_id: int = int(data['discount_id'])
        self.discount: Discount = Discount(data['discount'])

    def __repr__(self) -> str:
        return f'<DiscountOffer id={self.id} discount={self.discount!r}>'

    def is_acked(self) -> bool:
        """:class:`bool`: Checks if the discount offer has been acknowledged."""
        return self.expires_at is not None

    async def ack(self) -> None:
        """|coro|

        Acknowledges the discount offer.

        Raises
        ------
        HTTPException
            Acknowledging the discount offer failed.
        """
        data = await self._state.http.ack_user_offer(discount_offer_id=self.id)
        if not data:
            return

        # The type checker has no idea what is going on here
        if data.get('user_discount_offer') and int(data['user_discount_offer']['id']) == self.id:  # type: ignore
            self._update(data['user_discount_offer'])  # type: ignore
        if data.get('user_discount') and int(data['user_discount']['id']) == self.id:  # type: ignore
            self._update(data['user_discount'])  # type: ignore

    async def redeem(self) -> List[DiscountOffer]:
        """|coro|

        Applies the discount on the user's existing subscription.

        Raises
        ------
        NotFound
            The discount offer was not found.
        HTTPException
            Redeeming the discount offer failed.

        Returns
        -------
        List[:class:`DiscountOffer`]
            The applied discount offers.
        """
        data = await self._state.http.redeem_user_offer(self.id)
        return [DiscountOffer(data=item, state=self._state) for item in data]


class Discount(Hashable):
    """Represents a Discord subscription discount.

    .. container:: operations

        .. describe:: x != y

            Checks if two discounts are not equal.

        .. describe:: hash(x)

            Returns the discount's hash.

    .. versionadded:: 2.1

    Attributes
    ----------
    id: :class:`int`
        The ID of the discount.
    amount: :class:`int`
        The amount of the discount.
    created_at: :class:`datetime.datetime`
        When the discount was created.
    starts_at: Optional[:class:`datetime.datetime`]
        When the discount starts.
    ends_at: Optional[:class:`datetime.datetime`]
        When the discount ends.
    subscription_plan_ids: List[:class:`int`]
        The IDs of the subscription plans the discount applies to.
    sku_group_ids: List[:class:`int`]
        The IDs of the SKU groups the discount applies to.
    sku_ids: List[:class:`int`]
        The IDs of the SKUs the discount applies to.
    usage_limit: :class:`int`
        How many times the user can use the discount.
    interval: :class:`SubscriptionInterval`
        The interval of the discount.
    interval_count: :class:`int`
        The number of intervals in one usage of the discount.
    """

    __slots__ = (
        'id',
        'amount',
        'created_at',
        'starts_at',
        'ends_at',
        'subscription_plan_ids',
        'sku_group_ids',
        'sku_ids',
        'usage_limit',
        'interval',
        'interval_count',
    )

    def __init__(self, data: DiscountPayload) -> None:
        self.id: int = int(data['id'])
        self.amount: int = data['amount']
        self.created_at: datetime = parse_time(data['created_at'])
        self.starts_at: Optional[datetime] = parse_time(data.get('starts_at'))
        self.ends_at: Optional[datetime] = parse_time(data.get('ends_at'))
        self.subscription_plan_ids: List[int] = [int(plan_id) for plan_id in data.get('plan_ids') or []]
        self.sku_group_ids: List[int] = [int(sku_group_id) for sku_group_id in data.get('sku_group_ids') or []]
        self.sku_ids: List[int] = [int(sku_id) for sku_id in data.get('sku_ids') or []]
        self.usage_limit: int = data['user_usage_limit']
        self.interval: SubscriptionInterval = try_enum(SubscriptionInterval, data['user_usage_limit_interval'])
        self.interval_count: int = data['user_usage_limit_interval_count']

    def __repr__(self) -> str:
        return f'<Discount id={self.id} amount={self.amount} starts_at={self.starts_at!r} ends_at={self.ends_at!r}>'

    @property
    def duration(self) -> timedelta:
        """:class:`datetime.timedelta`: How long one usage of the discount lasts."""
        return timedelta(days=self.interval_count * self.interval.duration)


class PricingPromotion:
    """Represents a Discord localized pricing promotion.

    .. versionadded:: 2.0

    Attributes
    ----------
    subscription_plan_id: :class:`int`
        The ID of the subscription plan the promotion is for.
    country_code: :class:`str`
        The country code the promotion applies to.
    payment_source_types: List[:class:`PaymentSourceType`]
        The payment source types the promotion is restricted to.
    amount: :class:`int`
        The discounted price of the subscription plan.
    currency: :class:`str`
        The currency of the discounted price.
    """

    __slots__ = (
        'subscription_plan_id',
        'country_code',
        'payment_source_types',
        'amount',
        'currency',
    )

    def __init__(self, *, data: PricingPromotionPayload) -> None:
        self.subscription_plan_id: int = int(data['plan_id'])
        self.country_code: str = data['country_code']
        self.payment_source_types: List[PaymentSourceType] = [
            try_enum(PaymentSourceType, t) for t in data['payment_source_types']
        ]

        price = data['price']
        self.amount: int = price['amount']
        self.currency: str = price['currency']

    def __repr__(self) -> str:
        return f'<PricingPromotion plan_id={self.subscription_plan_id} country_code={self.country_code!r} amount={self.amount} currency={self.currency!r}>'



"""
The MIT License (MIT)

Copyright (c) 2021-present Dolfies

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
"""

from __future__ import annotations

from typing import TYPE_CHECKING, Optional, Union

from .enums import DirectoryCategory, DirectoryEntryType, try_enum
from .scheduled_event import ScheduledEvent
from .utils import MISSING, parse_time

if TYPE_CHECKING:
    from datetime import datetime

    from .channel import DirectoryChannel
    from .guild import Guild
    from .member import Member
    from .state import ConnectionState
    from .types.directory import (
        DirectoryEntry as DirectoryEntryPayload,
        PartialDirectoryEntry as PartialDirectoryEntryPayload,
    )

__all__ = ('DirectoryEntry',)


class DirectoryEntry:
    """Represents a directory entry for a channel.

    .. container:: operations

        .. describe:: x == y

            Checks if two entries are equal.

        .. describe:: x != y

            Checks if two entries are not equal.

        .. describe:: hash(x)

            Returns the entry's hash.

    .. versionadded:: 2.1

    Attributes
    -----------
    channel: :class:`DirectoryChannel`
        The channel this entry is from.
    type: :class:`DirectoryEntryType`
        The type of this entry.
    category: :class:`DirectoryCategory`
        The primary category of this entry.
    author_id: :class:`int`
        The ID of the user who created this entry.
    created_at: :class:`datetime.datetime`
        When this entry was created.
    description: Optional[:class:`str`]
        The description of the entry's guild.
        Only applicable for entries of type :attr:`DirectoryEntryType.guild`.
    entity_id: :class:`int`
        The ID of the entity this entry represents.
    guild: Optional[:class:`Guild`]
        The guild this entry represents.
        For entries of type :attr:`DirectoryEntryType.scheduled_event`,
        this is the guild the scheduled event is from.
        Not available in all contexts.
    featurable: :class:`bool`
        Whether this entry's guild can be featured in the directory.
        Only applicable for entries of type :attr:`DirectoryEntryType.guild`.
    scheduled_event: Optional[:class:`ScheduledEvent`]
        The scheduled event this entry represents.
        Only applicable for entries of type :attr:`DirectoryEntryType.scheduled_event`.
    rsvp: :class:`bool`
        Whether the current user has RSVP'd to the scheduled event.
        Only applicable for entries of type :attr:`DirectoryEntryType.scheduled_event`.
    """

    def __init__(
        self,
        *,
        data: Union[DirectoryEntryPayload, PartialDirectoryEntryPayload],
        state: ConnectionState,
        channel: DirectoryChannel,
    ):
        self.channel = channel
        self._state = state
        self._update(data)

    def __repr__(self) -> str:
        return f'<DirectoryEntry channel={self.channel!r} type={self.type!r} category={self.category!r} author_id={self.author_id!r} guild={self.guild!r}>'

    def __hash__(self) -> int:
        return hash((self.channel.id, self.entity_id))

    def __eq__(self, other: object) -> bool:
        if isinstance(other, DirectoryEntry):
            return self.channel == other.channel and self.entity_id == other.entity_id
        return NotImplemented

    def _update(self, data: Union[DirectoryEntryPayload, PartialDirectoryEntryPayload]):
        state = self._state
        self.type: DirectoryEntryType = try_enum(DirectoryEntryType, data['type'])
        self.category: DirectoryCategory = try_enum(DirectoryCategory, data.get('primary_category_id', 0))
        self.author_id: int = int(data['author_id'])
        self.created_at: datetime = parse_time(data['created_at'])
        self.description: Optional[str] = data.get('description') or None
        self.entity_id: int = int(data['entity_id'])

        guild_data = data.get('guild', data.get('guild_scheduled_event', {}).get('guild'))
        self.guild: Optional[Guild] = state.create_guild(guild_data) if guild_data is not None else None
        self.featurable: bool = guild_data.get('featurable_in_directory', False) if guild_data is not None else False

        event_data = data.get('guild_scheduled_event')
        self.scheduled_event: Optional[ScheduledEvent] = (
            ScheduledEvent(data=event_data, state=state) if event_data is not None else None
        )
        self.rsvp: bool = event_data.get('user_rsvp', False) if event_data is not None else False

    @property
    def author(self) -> Optional[Member]:
        """Optional[:class:`Member`]: The member that created this entry."""
        return self.channel.guild.get_member(self.author_id)

    async def edit(self, *, description: Optional[str] = MISSING, category: DirectoryCategory = MISSING) -> None:
        """|coro|

        Edits this directory entry.
        Only entries of type :attr:`DirectoryEntryType.guild` can be edited.

        You must be the author of the entry or have
        :attr:`~Permissions.manage_guild` in the represented guild to edit it.

        Parameters
        -----------
        description: Optional[:class:`str`]
            The new description of the entry's guild.
        category: :class:`DirectoryCategory`
            The new primary category of the entry.

        Raises
        -------
        Forbidden
            You do not have permissions to edit this entry.
        HTTPException
            Editing the entry failed.
        """
        data = await self._state.http.edit_directory_entry(
            self.channel.id,
            self.entity_id,
            description=description,
            primary_category_id=category.value if category is not MISSING else MISSING,
        )
        self._update(data)

    async def delete(self) -> None:
        """|coro|

        Deletes this directory entry.

        You must be the author of the entry or have
        :attr:`~Permissions.manage_guild` in the represented guild to delete it.

        Raises
        -------
        Forbidden
            You do not have permissions to delete this entry.
        HTTPException
            Deleting the entry failed.
        """
        await self._state.http.delete_directory_entry(self.channel.id, self.entity_id)



"""
The MIT License (MIT)

Copyright (c) 2015-present Rapptz

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
"""

from __future__ import annotations

from typing import List, Literal, Optional, TypedDict, Union
from typing_extensions import NotRequired, Required

from .snowflake import Snowflake, SnowflakeList
from .member import Member, UserWithMember
from .user import PartialUser
from .emoji import PartialEmoji
from .embed import Embed
from .channel import ChannelType
from .components import MessageActionRow
from .interactions import MessageInteraction
from .application import BaseApplication, PartialApplication
from .sticker import StickerItem
from .threads import Thread, ThreadMember
from .poll import Poll


class PartialMessage(TypedDict):
    channel_id: Snowflake
    guild_id: NotRequired[Snowflake]


class ChannelMention(TypedDict):
    id: Snowflake
    guild_id: Snowflake
    type: ChannelType
    name: str


class ReactionCountDetails(TypedDict):
    burst: int
    normal: int


ReactionType = Literal[0, 1]


class Reaction(TypedDict):
    count: int
    me: bool
    emoji: PartialEmoji
    me_burst: bool
    count_details: ReactionCountDetails
    burst_colors: List[str]


class Attachment(TypedDict):
    id: Snowflake
    filename: str
    size: int
    url: str
    proxy_url: str
    title: NotRequired[str]
    height: NotRequired[Optional[int]]
    width: NotRequired[Optional[int]]
    description: NotRequired[str]
    content_type: NotRequired[str]
    ephemeral: NotRequired[bool]
    duration_secs: NotRequired[float]
    waveform: NotRequired[str]
    flags: NotRequired[int]
    placeholder_version: NotRequired[int]
    placeholder: NotRequired[str]
    clip_created_at: NotRequired[str]
    clip_participants: NotRequired[List[PartialUser]]
    application: NotRequired[PartialApplication]


MessageActivityType = Literal[1, 2, 3, 5]


class MessageActivity(TypedDict):
    type: MessageActivityType
    party_id: str


MessageReferenceType = Literal[0, 1]


class MessageReference(TypedDict, total=False):
    type: MessageReferenceType
    message_id: Snowflake
    channel_id: Required[Snowflake]
    guild_id: Snowflake
    fail_if_not_exists: bool


class Call(TypedDict):
    participants: List[Snowflake]
    ended_timestamp: Optional[str]


class RoleSubscriptionData(TypedDict):
    role_subscription_listing_id: Snowflake
    tier_name: str
    total_months_subscribed: int
    is_renewal: bool


PurchaseNotificationResponseType = Literal[0]


class GuildProductPurchase(TypedDict):
    listing_id: Snowflake
    product_name: str


class PurchaseNotificationResponse(TypedDict):
    type: PurchaseNotificationResponseType
    guild_product_purchase: Optional[GuildProductPurchase]


MessageType = Literal[
    0,
    1,
    2,
    3,
    4,
    5,
    6,
    7,
    8,
    9,
    10,
    11,
    12,
    14,
    15,
    18,
    19,
    20,
    21,
    22,
    23,
    24,
    25,
    26,
    27,
    28,
    29,
    30,
    31,
    32,
    36,
    37,
    38,
    39,
    44,
    46,
    63,
]


class MessageSnapshot(TypedDict):
    type: MessageType
    content: str
    embeds: List[Embed]
    attachments: List[Attachment]
    timestamp: str
    edited_timestamp: Optional[str]
    flags: NotRequired[int]
    mentions: List[UserWithMember]
    mention_roles: SnowflakeList
    sticker_items: NotRequired[List[StickerItem]]
    components: NotRequired[List[MessageActionRow]]


class Message(PartialMessage):
    id: Snowflake
    author: PartialUser
    content: str
    timestamp: str
    edited_timestamp: Optional[str]
    tts: bool
    mention_everyone: bool
    mentions: List[UserWithMember]
    mention_roles: SnowflakeList
    attachments: List[Attachment]
    embeds: List[Embed]
    pinned: bool
    poll: NotRequired[Poll]
    type: MessageType
    member: NotRequired[Member]
    mention_channels: NotRequired[List[ChannelMention]]
    reactions: NotRequired[List[Reaction]]
    nonce: NotRequired[Union[int, str]]
    webhook_id: NotRequired[Snowflake]
    activity: NotRequired[MessageActivity]
    application: NotRequired[BaseApplication]
    application_id: NotRequired[Snowflake]
    message_reference: NotRequired[MessageReference]
    flags: NotRequired[int]
    sticker_items: NotRequired[List[StickerItem]]
    referenced_message: NotRequired[Optional[Message]]
    interaction: NotRequired[MessageInteraction]
    components: NotRequired[List[MessageActionRow]]
    position: NotRequired[int]
    call: NotRequired[Call]
    role_subscription_data: NotRequired[RoleSubscriptionData]
    hit: NotRequired[bool]
    thread: NotRequired[Thread]
    purchase_notification: NotRequired[PurchaseNotificationResponse]


AllowedMentionType = Literal['roles', 'users', 'everyone']


class AllowedMentions(TypedDict):
    parse: List[AllowedMentionType]
    roles: SnowflakeList
    users: SnowflakeList
    replied_user: bool


class MessageSearchIndexingResult(TypedDict):
    # Error but not quite
    message: str
    code: int
    documents_indexed: int
    retry_after: int


class MessageSearchResult(TypedDict):
    messages: List[List[Message]]
    threads: NotRequired[List[Thread]]
    members: NotRequired[List[ThreadMember]]
    total_results: int
    analytics_id: str
    doing_deep_historical_index: NotRequired[bool]


MessageSearchAuthorType = Literal['user', '-user', 'bot', '-bot', 'webhook', '-webhook']
MessageSearchHasType = Literal[
    'image',
    '-image',
    'sound',
    '-sound',
    'video',
    '-video',
    'file',
    '-file',
    'sticker',
    '-sticker',
    'embed',
    '-embed',
    'link',
    '-link',
]
MessageSearchSortType = Literal['timestamp', 'relevance']
MessageSearchSortOrder = Literal['desc', 'asc']


class PartialAttachment(TypedDict, total=False):
    id: Snowflake
    filename: Required[str]
    description: str
    uploaded_filename: str
    title: str
    duration_secs: float
    waveform: str
    is_spoiler: bool
    is_remix: bool
    is_clip: bool
    application_id: Snowflake
    clip_created_at: str
    clip_participant_ids: SnowflakeList


class UploadedAttachment(TypedDict):
    id: NotRequired[Snowflake]
    filename: str
    file_size: int
    is_clip: NotRequired[bool]


class CloudAttachment(TypedDict):
    id: NotRequired[Snowflake]
    upload_url: str
    upload_filename: str


class CloudAttachments(TypedDict):
    attachments: List[CloudAttachment]


class TypingResponse(TypedDict):
    message_send_cooldown_ms: int
    thread_create_cooldown_ms: int


class MessagePin(TypedDict):
    pinned_at: str
    message: Message


class ChannelPins(TypedDict):
    items: List[MessagePin]
    has_more: bool



"""
The MIT License (MIT)

Copyright (c) 2021-present Dolfies

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
"""

from __future__ import annotations

from typing import TYPE_CHECKING, List, Sequence

from .utils import SequenceProxy

if TYPE_CHECKING:
    from typing_extensions import Self

    from .state import ConnectionState
    from .types.gateway import Tutorial as TutorialPayload

# fmt: off
__all__ = (
    'Tutorial',
)
# fmt: on


class Tutorial:
    """Represents the Discord new user tutorial state.

    .. versionadded:: 2.1

    Attributes
    -----------
    suppressed: :class:`bool`
        Whether the tutorial is suppressed or not.
    """

    __slots__ = ('suppressed', '_indicators', '_state')

    def __init__(self, *, data: TutorialPayload, state: ConnectionState):
        self._state: ConnectionState = state
        self.suppressed: bool = data.get('indicators_suppressed', True)
        self._indicators: List[str] = data.get('indicators_confirmed', [])

    def __repr__(self) -> str:
        return f'<Tutorial suppressed={self.suppressed} indicators={self._indicators!r}>'

    @classmethod
    def default(cls, state: ConnectionState) -> Self:
        self = cls.__new__(cls)
        self._state = state
        self.suppressed = True
        self._indicators = []
        return self

    @property
    def indicators(self) -> Sequence[str]:
        """Sequence[:class:`str`]: A list of the tutorial indicators that have been confirmed."""
        return SequenceProxy(self._indicators)

    async def suppress(self) -> None:
        """|coro|

        Permanently suppresses all tutorial indicators.

        Raises
        -------
        HTTPException
            Suppressing the tutorial failed.
        """
        await self._state.http.suppress_tutorial()
        self.suppressed = True

    async def confirm(self, *indicators: str) -> None:
        r"""|coro|

        Confirms a list of tutorial indicators.

        Parameters
        -----------
        \*indicators: :class:`str`
            The indicators to confirm.

        Raises
        -------
        HTTPException
            Confirming the tutorial indicators failed.
        """
        req = self._state.http.confirm_tutorial_indicator
        # The gateway does not send updates on the tutorial
        # So we keep the state updated ourselves
        for indicator in indicators:
            if indicator not in self.indicators:
                await req(indicator)
                self._indicators.append(indicator)

        # Indicators are sorted alphabetically
        self._indicators.sort()



"""
The MIT License (MIT)

Copyright (c) 2021-present Dolfies

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
"""

from __future__ import annotations

from typing import TYPE_CHECKING, List, Optional

from .application import PartialApplication
from .guild import UserGuild
from .mixins import Hashable
from .utils import MISSING

if TYPE_CHECKING:
    from .abc import Snowflake
    from .permissions import Permissions
    from .state import ConnectionState
    from .user import User
    from .types.oauth2 import OAuth2Authorization as OAuth2AuthorizationPayload, OAuth2Token as OAuth2TokenPayload

__all__ = (
    'OAuth2Token',
    'OAuth2Authorization',
)


class OAuth2Token(Hashable):
    """Represents an authorized OAuth2 application for a user.

    .. container:: operations

        .. describe:: x == y

            Checks if two authorizations are equal.

        .. describe:: x != y

            Checks if two authorizations are not equal.

        .. describe:: hash(x)

            Return the authorizations's hash.

        .. describe:: str(x)

            Returns the authorizations's name.

    .. versionadded:: 2.1

    Attributes
    -----------
    id: :class:`int`
        The ID of the authorization.
    application: :class:`PartialApplication`
        The application that the authorization is for.
    scopes: List[:class:`str`]
        The scopes that the authorization has.
    """

    __slots__ = ('id', 'application', 'scopes', '_state')

    def __init__(self, *, state: ConnectionState, data: OAuth2TokenPayload):
        self._state = state
        self.id: int = int(data['id'])
        self.application: PartialApplication = PartialApplication(state=state, data=data['application'])
        self.scopes: List[str] = data['scopes']

    def __repr__(self):
        return f'<OAuth2Token id={self.id} application={self.application!r} scopes={self.scopes!r}>'

    def __str__(self):
        return self.application.name

    @property
    def authorized(self) -> bool:
        """:class:`bool`: Whether the user has already authorized the application.

        This is here for compatibility purposes and is always ``True``.
        """
        return True

    async def revoke(self):
        """|coro|

        Revokes the application's authorization.

        Raises
        -------
        HTTPException
            Deauthorizing the application failed.
        """
        await self._state.http.revoke_oauth2_token(self.id)


class OAuth2Authorization:
    """Represents a Discord OAuth2 application authorization.

    .. versionadded:: 2.1

    Attributes
    -----------
    scopes: List[:class:`str`]
        The scopes that the authorization has.
    response_type: Optional[:class:`str`]
        The response type that will be used for the authorization, if using the full OAuth2 flow.
    code_challenge_method: Optional[:class:`str`]
        The code challenge method that will be used for the PKCE authorization, if using the full OAuth2 flow.
    code_challenge: Optional[:class:`str`]
        The code challenge that will be used for the PKCE authorization, if using the full OAuth2 flow.
    state: Optional[:class:`str`]
        The state that will be used for authorization security.
    authorized: :class:`bool`
        Whether the user has already authorized the application.
    application: :class:`PartialApplication`
        The application that the authorization is for.
    bot: Optional[:class:`User`]
        The bot user associated with the application, provided if authorizing with the ``bot`` scope.
    approximate_guild_count: Optional[:class:`int`]
        The approximate number of guilds the bot is in, provided if authorizing with the ``bot`` scope.
    guilds: List[:class:`UserGuild`]
        The guilds the current user is in, provided if authorizing with the ``bot`` scope.
    redirect_uri: Optional[:class:`str`]
        The redirect URI that will be used for the authorization, if using the full OAuth2 flow and a redirect URI exists.
    """

    __slots__ = (
        'authorized',
        'application',
        'bot',
        'approximate_guild_count',
        'guilds',
        'redirect_uri',
        '_original_redirect_uri',
        'scopes',
        'response_type',
        'code_challenge_method',
        'code_challenge',
        'state',
        'nonce',
        '_state',
    )

    def __init__(
        self,
        *,
        _state: ConnectionState,
        data: OAuth2AuthorizationPayload,
        scopes: List[str],
        response_type: Optional[str] = None,
        original_redirect_uri: Optional[str] = None,
        code_challenge_method: Optional[str] = None,
        code_challenge: Optional[str] = None,
        state: Optional[str] = None,
        nonce: Optional[str] = None,
    ):
        self._state = _state
        self.scopes: List[str] = scopes
        self.response_type: Optional[str] = response_type
        self._original_redirect_uri: Optional[str] = original_redirect_uri
        self.code_challenge_method: Optional[str] = code_challenge_method
        self.code_challenge: Optional[str] = code_challenge
        self.state: Optional[str] = state
        self.nonce: Optional[str] = nonce
        self.authorized: bool = data['authorized']
        self.application: PartialApplication = PartialApplication(state=_state, data=data['application'])
        self.bot: Optional[User] = _state.store_user(data['bot']) if 'bot' in data else None
        self.approximate_guild_count: Optional[int] = (
            data['bot'].get('approximate_guild_count', 0) if 'bot' in data else None
        )
        self.guilds: List[UserGuild] = [UserGuild(state=_state, data=g) for g in data.get('guilds', [])]
        self.redirect_uri: Optional[str] = data.get('redirect_uri')

    def __repr__(self):
        return f'<OAuth2Authorization authorized={self.authorized} application={self.application!r} scopes={self.scopes!r} response_type={self.response_type!r} redirect_uri={self.redirect_uri}>'

    async def authorize(
        self, *, guild: Snowflake = MISSING, channel: Snowflake = MISSING, permissions: Permissions = MISSING
    ) -> str:
        """|coro|

        Authorizes the application for the user. A shortcut for :meth:`Client.create_authorization`.

        Parameters
        -----------
        guild: :class:`Guild`
            The guild to authorize for, if authorizing with the ``applications.commands`` or ``bot`` scopes.
        channel: Union[:class:`TextChannel`, :class:`VoiceChannel`, :class:`StageChannel`]
            The channel to authorize for, if authorizing with the ``webhooks.incoming`` scope. See :meth:`Guild.webhook_channels`.
        permissions: :class:`Permissions`
            The permissions to grant, if authorizing with the ``bot`` scope.

        Raises
        -------
        HTTPException
            Authorizing the application failed.

        Returns
        --------
        :class:`str`
            The URL to redirect the user to. May be an error page.
        """
        data = await self._state.http.authorize_oauth2(
            self.application.id,
            self.scopes,
            self.response_type,
            self._original_redirect_uri,
            self.code_challenge_method,
            self.code_challenge,
            self.state,
            self.nonce,
            guild_id=guild.id if guild else None,
            webhook_channel_id=channel.id if channel else None,
            permissions=permissions.value if permissions else None,
        )
        return data['location']



"""
The MIT License (MIT)

Copyright (c) 2015-present Rapptz

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
"""

from __future__ import annotations

from typing import Any, Dict, List, Literal, Optional, TypedDict, Union
from typing_extensions import NotRequired
from .user import PartialUser
from .snowflake import Snowflake


StatusType = Literal['idle', 'dnd', 'online', 'offline']
StatusDisplayType = Literal[0, 1, 2]


class UserID(TypedDict):
    id: Snowflake


class BasePresenceUpdate(TypedDict):
    status: StatusType
    activities: List[Activity]
    hidden_activities: NotRequired[List[Activity]]
    client_status: ClientStatus
    processed_at_timestamp: Union[Literal[0], str]


class UserPresenceUpdate(BasePresenceUpdate):
    user: PartialUser


class PartialPresenceUpdate(BasePresenceUpdate):
    user: Union[PartialUser, UserID]
    guild_id: NotRequired[Snowflake]


class ClientStatus(TypedDict, total=False):
    desktop: StatusType
    mobile: StatusType
    web: StatusType
    embedded: StatusType


class ActivityTimestamps(TypedDict, total=False):
    start: int
    end: int


class ActivityParty(TypedDict, total=False):
    id: str
    size: List[int]


class ActivityAssets(TypedDict, total=False):
    large_image: str
    large_text: str
    small_image: str
    small_text: str
    large_url: str
    small_url: str
    invite_cover_image: str


class ActivitySecrets(TypedDict, total=False):
    join: str
    spectate: str


class ActivitySecret(TypedDict):
    secret: str


class ActivityEmoji(TypedDict):
    name: Optional[str]
    id: Optional[Snowflake]
    animated: NotRequired[bool]


ActivityType = Literal[0, 1, 2, 3, 4, 5, 6]


class _BaseActivity(TypedDict):
    name: str
    type: ActivityType

    # Receive-only fields
    id: NotRequired[str]
    created_at: NotRequired[int]


class Activity(_BaseActivity, total=False):
    url: Optional[str]
    state: Optional[str]
    details: Optional[str]
    timestamps: ActivityTimestamps
    platform: Optional[str]
    supported_platforms: List[str]
    assets: ActivityAssets
    party: ActivityParty
    application_id: Snowflake
    parent_application_id: Snowflake
    flags: int
    emoji: ActivityEmoji
    secrets: ActivitySecrets
    metadata: NotRequired[Dict[str, Any]]
    session_id: Optional[str]
    instance: bool
    buttons: List[str]
    sync_id: str
    state_url: str
    details_url: str
    status_display_type: Optional[StatusDisplayType]


HangStatusVariantType = Literal['illocons', 'twemoji', 'twemojimild']


class SettingsActivity(TypedDict, total=False):
    text: str
    emoji_id: Snowflake
    emoji_name: str
    expires_at: str



"""
The MIT License (MIT)

Copyright (c) 2021-present Dolfies

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
"""

from __future__ import annotations

from typing import List, Literal, Optional, Tuple, TypedDict, Union

from typing_extensions import NotRequired


class ExperimentResponse(TypedDict):
    fingerprint: NotRequired[str]
    assignments: List[UserExperiment]


class ExperimentResponseWithGuild(ExperimentResponse):
    guild_experiments: NotRequired[List[GuildExperiment]]


class RolloutData(TypedDict):
    s: int
    e: int


Rollout = Tuple[int, List[RolloutData]]


Filters = List[
    Union[
        Tuple[Literal[1604612045], Tuple[Tuple[Literal[1183251248], List[str]]]],  # FEATURE
        Tuple[
            Literal[2404720969], Tuple[Tuple[Literal[3399957344], Optional[int]], Tuple[Literal[1238858341], int]]
        ],  # ID_RANGE
        Tuple[
            Literal[2918402255], Tuple[Tuple[Literal[3399957344], Optional[int]], Tuple[Literal[1238858341], int]]
        ],  # MEMBER_COUNT_RANGE
        Tuple[Literal[3013771838], Tuple[Tuple[Literal[3013771838], List[int]]]],  # IDs
        Tuple[Literal[4148745523], Tuple[Tuple[Literal[4148745523], List[int]]]],  # HUB_TYPE
        Tuple[Literal[188952590], Tuple[Tuple[Literal[188952590], bool]]],  # VANITY_URL
        Tuple[Literal[2294888943], Tuple[Tuple[Literal[2690752156], int], Tuple[Literal[1982804121], int]]],  # RANGE_BY_HASH
    ]
]


Population = Tuple[
    List[Rollout],  # rollouts
    Filters,  # filters
]


class Override(TypedDict):
    b: int
    k: List[int]


Holdout = Tuple[
    int,  # bucket
    str,  # experiment_name
]


UserExperiment = Tuple[
    int,  # hash
    int,  # revision
    int,  # bucket
    Literal[-1, 0],  # override
    int,  # population
    int,  # hash_result
    Literal[0, 1],  # aa_mode
    Literal[0, 1],  # trigger_debugging
    Optional[str],  # holdout_name
    Optional[int],  # holdout_revision
    Optional[int],  # holdout_bucket
]


GuildExperiment = Tuple[
    int,  # hash
    Optional[str],  # hash_key
    int,  # revision
    List[Population],  # populations
    List[Override],  # overrides
    List[List[Population]],  # overrides_formatted
    Optional[str],  # holdout_name
    Optional[int],  # holdout_bucket
    Literal[0, 1],  # aa_mode
    Literal[0, 1],  # trigger_debugging
]
